using System;
using System.Windows.Forms;
using System.Runtime.InteropServices;

// this is an intermediate class to fix the undo behavior
// undo doesn't work when calling SelectedText.Length before
// calling undo

[StructLayout(LayoutKind.Sequential)]
class CHARRANGE
{
	public Int32 cpMin;
	public Int32 cpMax;
}

[StructLayout(LayoutKind.Sequential)]
class GETTEXTLENGTHEX
{
	public UInt32 flags;
	public UInt32 codepage;
}

[StructLayout(LayoutKind.Sequential, CharSet=CharSet.Ansi)]
class GETTEXTEX
{
	public UInt32 cb;
	public UInt32 flags;
	public UInt32 codepage;
	public String lpDefaultChar;	// decorate with ANSI -- ignored (NULL)
	public IntPtr lpUsedDefChar;	// NULL
}

[StructLayout(LayoutKind.Sequential)]
class SETTEXTEX
{
	public UInt32 flags;
	public UInt32 codepage;
}

struct User32
{
	public const Int32 WM_USER = 0x0400;

	public const Int32 EM_EXGETSEL = WM_USER + 52;
	public const Int32 EM_EXSETSEL = WM_USER + 55;
	public const Int32 EM_GETTEXTEX = WM_USER + 94;
	public const Int32 EM_GETTEXTLENGTHEX = WM_USER + 95;
	public const Int32 EM_SETTEXTEX = WM_USER + 97;

	public const Int32 ST_DEFAULT = 0;
	public const Int32 ST_KEEPUNDO = 1;

	public const Int32 GT_DEFAULT = 0;

	public const Int32 GTL_DEFAULT = 0;

	[DllImport("user32.dll", EntryPoint="SendMessage", CharSet=CharSet.Auto)]
	public static extern int SendMessage(IntPtr hWnd, Int32 Msg, IntPtr wParam, IntPtr lParam);
}

class RichTextBoxBase : RichTextBox
{
	CHARRANGE _CharRange;
	IntPtr _CharRangePtr = IntPtr.Zero;
	GETTEXTLENGTHEX _GetTextLengthEx;
	IntPtr _GetTextLengthExPtr = IntPtr.Zero;
	GETTEXTEX _GetTextEx;
	IntPtr _GetTextExPtr = IntPtr.Zero;
	SETTEXTEX _SetTextEx;
	IntPtr _SetTextExPtr = IntPtr.Zero;

	public RichTextBoxBase()
	{
		_CharRange = new CHARRANGE();
		_GetTextLengthEx = new GETTEXTLENGTHEX();
		_GetTextEx = new GETTEXTEX();
		_SetTextEx = new SETTEXTEX();

		_CharRangePtr = Marshal.AllocHGlobal(Marshal.SizeOf(_CharRange));

		_GetTextLengthExPtr = Marshal.AllocHGlobal(Marshal.SizeOf(_GetTextLengthEx));
		_GetTextLengthEx.flags = User32.GTL_DEFAULT;
		_GetTextLengthEx.codepage = 1200;	// unicode
		Marshal.StructureToPtr(_GetTextLengthEx, _GetTextLengthExPtr, false);

		_GetTextExPtr = Marshal.AllocHGlobal(Marshal.SizeOf(_GetTextEx));

		_SetTextExPtr = Marshal.AllocHGlobal(Marshal.SizeOf(_SetTextEx));
	}

	~RichTextBoxBase()
	{
		Dispose(false);
	}

	/*public void Dispose()
	{
		GC.SuppressFinalize(this);
		Dispose(true);
	}*/

	protected override void Dispose(bool disposing)
	{
		if (_CharRangePtr != IntPtr.Zero)
			Marshal.FreeHGlobal(_CharRangePtr);
		if (_GetTextLengthExPtr != IntPtr.Zero)
			Marshal.FreeHGlobal(_GetTextLengthExPtr);
		if (_GetTextExPtr != IntPtr.Zero)
			Marshal.FreeHGlobal(_GetTextExPtr);
		if (_SetTextExPtr != IntPtr.Zero)
			Marshal.FreeHGlobal(_SetTextExPtr);
		if (disposing)
		{
			// TODO: If the class has members that wrap
			// unmanaged resources, call Close or Dispose on
			// them here.
		}
	}

	public Int32 GetSelectedTextLength()
	{
		_CharRange.cpMin = 0;
		_CharRange.cpMax = 0;
		Marshal.StructureToPtr(_CharRange, _CharRangePtr, false);
		User32.SendMessage(this.Handle, User32.EM_EXGETSEL, IntPtr.Zero, _CharRangePtr);
		Marshal.PtrToStructure(_CharRangePtr, _CharRange);

		return _CharRange.cpMax - _CharRange.cpMin;
	}

	public Int32 GetTextLength()
	{
		return User32.SendMessage(this.Handle, User32.EM_GETTEXTLENGTHEX,
								  _GetTextLengthExPtr, IntPtr.Zero);
	}

	public void SelectAllText()
	{
		_CharRange.cpMin = 0;
		_CharRange.cpMax = -1;
		Marshal.StructureToPtr(_CharRange, _CharRangePtr, false);
		User32.SendMessage(this.Handle, User32.EM_EXSETSEL, IntPtr.Zero,
						   _CharRangePtr);
	}

	public String GetText()
	{
		IntPtr buffer = IntPtr.Zero;
		try
		{
			const Int32 charsize = 2;

			_GetTextEx.cb = (UInt32)((GetTextLength() + 1) * charsize);
			_GetTextEx.flags = User32.GT_DEFAULT;
			_GetTextEx.codepage = 1200;
			_GetTextEx.lpDefaultChar = null;
			_GetTextEx.lpUsedDefChar = IntPtr.Zero;

			Marshal.StructureToPtr(_GetTextEx, _GetTextExPtr, false);

			buffer = Marshal.AllocHGlobal((Int32)_GetTextEx.cb);
			Int32 x = User32.SendMessage(this.Handle, User32.EM_GETTEXTEX, _GetTextExPtr,
							   			 buffer);
			String text = Marshal.PtrToStringUni(buffer);
			return text;
		}
		finally
		{
			if (buffer != IntPtr.Zero)
				Marshal.FreeHGlobal(buffer);
		}
	}

	public void SetText(String text)
	{
		IntPtr ptr = IntPtr.Zero;
		try
		{
			_SetTextEx.flags = User32.ST_KEEPUNDO;
			_SetTextEx.codepage = 1200;
			Marshal.StructureToPtr(_SetTextEx, _SetTextExPtr, false);

			ptr = Marshal.StringToHGlobalUni(text);
			User32.SendMessage(this.Handle, User32.EM_SETTEXTEX, _SetTextExPtr, ptr);
		}
		finally
		{
			if (ptr != IntPtr.Zero)
				Marshal.FreeHGlobal(ptr);
		}
	}

	public void Reset()
	{
		IntPtr ptr = IntPtr.Zero;
		try
		{
			_SetTextEx.flags = User32.ST_DEFAULT;
			_SetTextEx.codepage = 1200;
			Marshal.StructureToPtr(_SetTextEx, _SetTextExPtr, false);

			ptr = Marshal.StringToHGlobalUni("");
			User32.SendMessage(this.Handle, User32.EM_SETTEXTEX, _SetTextExPtr, ptr);
		}
		finally
		{
			if (ptr != IntPtr.Zero)
				Marshal.FreeHGlobal(ptr);
		}
	}
}

